/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : sbuffer.cpp
  DESCRIPTION : the smart input buffer of sorting program (implementation)
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920906 AH
  LAST CHANGE : 950221 AH enough_memory_for was no accurate
  LAST CHANGE : 950220 AH enough_memory_for introduced
  LAST CHANGE : 950216 AH fixing of missing end separator
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950202 AH eof splitted into allread and eof
  LAST CHANGE : 950202 AH cleanup comment etc.
  LAST CHANGE : 950202 AH eof made false for remainderlen!=0
  LAST CHANGE : 950202 AH No unnecessary large merge buffers
  LAST CHANGE : 921112 AH comment clean up
  LAST CHANGE : 921109 AH TRACE introduced
  LAST CHANGE : 921108 AH References to stream used
  LAST CHANGE : 921106 AH Eliminated estimated length parameter on creation
  LAST CHANGE : 921106 AH Introduced CHAR typedef
  LAST CHANGE : 921106 AH Made ``Separators'' class
  LAST CHANGE : 921105 AH Only unsigned char's
  LAST CHANGE : 920926 Merge/SortBuffer specialisations of Buffer
  LAST CHANGE : 920916 cin and cout possible
                                                                   */
/*******************************************************************/


#include <alloc.h>
#include <string.h>
#include <assert.h>
#include <fstream.h>    // for ifstream, ofstream

#include "config.h"
#include "lib.h"
#include "sheap.h"
#include "search.h"
#include "sstring.h"
#include "sbuffer.h"

#define recordlengthestimate (Separators::recordlengthestimate)
#define maxfield (Separators::maxfield)

/****** Methods applicable to all buffers (class ``Buffer'') ****/

/*****************************************************************************/
/*                                                                           */
/*  Returns "There is enough memory on the heap to accommodate another       */
/*  buffer with the same size size as this one".                             */
/*  Use a safety margin of 10%                                               */
/*                                                                           */
/*****************************************************************************/

bool Buffer::enough_memory()
{
   return enough_memory_for( charbufsize, 1 );
}


/*****************************************************************************/
/*                                                                           */
/*  Returns "There is enough memory on the heap to accommodate ``amount''    */
/*  buffers with ``len'' characters each plus supporting buffers".           */
/*  Use a safety margin of 10%                                               */
/*                                                                           */
/*****************************************************************************/

bool Buffer::enough_memory_for( int len , int amount )
{
   // Number of records estimate, possibly limited by sheap
   // Prevent overflow in calculations involving it by making it long
   long numrec = min( 0xff00U/sizeof(SortString), len/recordlengthestimate+2 );

   // Memory needed for each character buffer plus structure information
   long weneed = len + numrec*( sizeof(SortString) +
                  (maxfield+1)*sizeof(int) );
   bool enough = 10*amount*weneed/9 < coreleft();

   TRACE<<"charbufsize            "     << len <<'\n';
   TRACE<<"recordlengthestimate"  <<recordlengthestimate<<'\n';
   TRACE<<"need: "<<weneed*amount<<"   have: "<<coreleft()<<'\n';
   TRACE<<"this is "<<(!enough?"not ":"")<<"enough\n";

   return enough;
}


/*****************************************************************************/
/*                                                                           */
/*  Create a buffer on the input stream ``sourcetoconnect'' to accommodate   */
/*  ``len'' characters. Allocate also the supporting structures for them,    */
/*  to best estimation, that will be needed for sorting it.                  */
/*  ``sourcetoconnect'' should already have been opened.                     */
/*                                                                           */
/*****************************************************************************/

Buffer::Buffer(istream &sourcetoconnect, int len ):
   SortedHeap< SortString > ( len/recordlengthestimate+2 )
{
   long int lenbufsize = heapsize*(maxfield+1L);

   // The following is a very seldom fenomenon:
   // occurs ONLY with records consisting of 4 one char fields
   if ( lenbufsize > 0x7f80 )
   {
        heapsize = 0x7F80U / (maxfield+1); // Restrict usable part of SortedHeap
        lenbufsize = heapsize*(maxfield+1L);
        TRACE << "Heapsize adjusted to "<<heapsize<<'\n';
   }

   source = &sourcetoconnect;
   lenbuf = new int[lenbufsize];
#ifdef DEBUG
   lenbufend = lenbuf + lenbufsize;
#endif
   charbuf = (new CHAR[len+1]);            // Extra 1 for ending zero or extra separator
   charbufsize = len;

   TRACE << "lenght of lenbuf :"<<lenbufsize <<'\n';
   TRACE << "lenght of charbuf :"<< len<<'\n';

   assert(lines);
   assert(lenbuf);
   assert(charbuf);
}


/*****************************************************************************/
/*                                                                           */
/*  ``getsortstring'' returns the next SortString from the unassigned part   */
/*  of the buffer: charbuf [bun..eun-1]. It finds the start of the next      */
/*  record and of its fields using the installed separator and returns a     */
/*  pointer to these lengths. See the definition of SortString. (Normally    */
/*  there is just one field and the lengths are zero and the record          */
/*  length.)                                                                 */
/*  If the buffer is exhausted, a null SortString is returned, i.e. one      */
/*  with a start field of null.                                              */
/*                                                                           */
/*****************************************************************************/

SortString Buffer::getsortstring()
{
   CHAR * start = charbuf+bun; // Start of the next SortString
   int *  plen = plenbuf;      // Length pointer of the next SortString
   int i;                      // Yet Another Counter

   CHAR * sepposition =  Separators::findseparator( start, eun-bun );
   nosep = !sepposition;

   // No more separator : no sortstrings
   if ( nosep )
      return SortString( 0, 0 );

   // Normal: separators okay fetch all field separators
   *plenbuf++ = 0; // First field starts at offset zero
   int len = sepposition-start; // FIrst field length
   assert ( 0<=len && len<=eun-bun );
   *plenbuf++ = len;
   // Now get other field lengths, if any
   for ( i = maxfield-1; i--; )
   {
      sepposition =  Separators::findseparator( start, eun-bun );
      len = sepposition-start;
      assert ( 0<=len && len<=eun-bun );
      *plenbuf++ = len;
   }
   assert( plenbuf <= lenbufend );

   if ( bun+len==eun && !allread )
   { // Do not yet accept the record, because there could be more matching
      plenbuf = plen;        //backtrack
      return SortString( 0, 0 );
   }
   else
   {
      bun += len;
      return SortString( start, plen );
   }
}


/*****************************************************************************/
/*                                                                           */
/*   ``sort'' chips off as many records as possible from the buffer          */
/*  ``charbuf'' and puts them in the SortedHeap. They are ordered by that    */
/*  very fact.                                                               */
/*                                                                           */
/*****************************************************************************/

void Buffer::sort()
{
   SortString ss;

   if ( eun == 0 ) return;

   charbuf[eun] = '\0';     // Important to prevent spurious string matches

   plenbuf = lenbuf;
   while ( !full() )
   {
      ss = getsortstring( );
      if ( !ss.start ) break;
      additem(ss);
   }

   // Remember, the last record of each buffer is rejected,
   // unless we are at the end of the input!
   if ( empty() && allread )
      if ( Separators::fixableC( charbuf, eun ) )
      {
         ss = getsortstring( );
         assert( ss.start );
         additem(ss);
         cerr<<"WARNING: Last record fixed by temporarily adding separator\n";
      }
      else
         cerr<<"Records do not correspond to description"<<amen;

   if ( empty() && charbufsize == eun )
         cerr<<"Records too big (>32K) or not according to description"<<amen;

   // In other cases of emptyness there is still progress to be made.

   if ( nlines>10 || nlines>0 && bun>charbufsize/3  )
      recordlengthestimate = bun/nlines;
   TRACE<<"New recordlengthestimate "<<recordlengthestimate<<'\n';
}


//#######################################################################
//##################  specific   ########################################
//##################    for      ########################################
//################## SortBuffer  ########################################
//#######################################################################

/*****************************************************************************/
/*                                                                           */
/*  Create a Buffer connected to ``sourcetoconnect'' with an input buffer    */
/*  of ``len'' characters and fill it from the current file position.        */
/*                                                                           */
/*****************************************************************************/

SortBuffer::SortBuffer(istream &sourcetoconnect, int len ):
Buffer( sourcetoconnect, len )
{
   refill();
}


/*****************************************************************************/
/*                                                                           */
/*  ``read'' reads the buffer ``charbuf'' full from the stream. It handles   */
/*  possible left overs from previous reads in order to make                 */
/*  charbuf[0..eun-1] contain the next (as yet unhandled) part of the        */
/*  input stream.                                                            */
/*                                                                           */
/*****************************************************************************/

void SortBuffer::read()
{
   bun = eun = 0;

   // Add `remainder'' to start of buffer
   if ( remainderstart )       // If it had been filled in
   {
      memmove( charbuf, remainderstart, remainderlen );
      eun += remainderlen;
   }

   // Fill from stream
   int toberead = charbufsize-eun;

   if ( !source->eof() )
   {
      source->read( charbuf+eun, toberead );
      if ( source->eof() )
         toberead = source->gcount();
      else
         assert( toberead == source->gcount() );
   }
   else
      toberead = 0;

   eun += toberead;

   allread = source->eof();
}


// Contains the left over from previous reads.
// It points into ``charbuf'' so be careful to copy the content,
// when it is added to a new ``charbuf''.
// It is essential that it is initialised to NULL (empty).
// This static initialisation is only sufficient if Sortbuffer is
// used only on one stream.
CHAR *SortBuffer::remainderstart = 0;
int SortBuffer::remainderlen;

/*****************************************************************************/
/*                                                                           */
/*  Read the buffer full from the current fileposition in                    */
/*  ``sourcetoconnect'' and transform it into a sorted heap of SortString    */
/*  's. Leftovers from the previous read are used up, before physically      */
/*  reading.                                                                 */
/*                                                                           */
/*****************************************************************************/

void SortBuffer::refill ( )
{
   read();
   sort();
   remainderstart = charbuf+bun;
   remainderlen   = eun-bun;

   // The outside world only cares whether the input file is
   // logically consumed, not whether it is physically read
   eof = source->eof() && 0==remainderlen;
}

//#######################################################################
//##################  specific   ########################################
//##################    for      ########################################
//################## MergeBuffer ########################################
//#######################################################################

/*****************************************************************************/
/*                                                                           */
/*  Create a Buffer connected to the part of a stream ``sourcetoconnect''    */
/*  starting at ``mergepoint[0]'' and ending just before ``mergepoint[1]'',  */
/*  with an input buffer of ``len'' characters. Fill it from the file        */
/*  position ``mergepoint[0]''.                                              */
/*                                                                           */
/*****************************************************************************/

MergeBuffer::MergeBuffer( ifstream &source, long *mergepoint, int len ):
// Buffer( source, min(len,mergepoint[1]-mergepoint[0]) )
Buffer( source, len )
{
   remainderstart = 0;
   remainderlen = 0;
   mergepointstart = mergepoint[0];
   mergepointend= mergepoint[1];
   refill();
}

/*****************************************************************************/
/*                                                                           */
/*  ``read'' reads the buffer full from part of the input file reserved for  */
/*  this mergebuffer. Possible left overs from previous reads are handled    */
/*  in order to make charbuf[0..eun-1] contains the next (as yet unhandled)  */
/*  part of the input stream.                                                */
/*                                                                           */
/*****************************************************************************/

void MergeBuffer::read()
{
   bun = eun = 0;

   // Add `remainder'' to start of buffer
   if ( remainderstart )       // If it had been filled in
   {
      memmove( charbuf, remainderstart, remainderlen );
      eun += remainderlen;
   }

   // Fill from stream
   int toberead = charbufsize-eun;
   if ( toberead>mergepointend-mergepointstart )  // Beware! longs
      toberead=mergepointend-mergepointstart;

   if ( toberead != 0 )
   {
      // Somewhat unexpectedly ``istream'' has allways a ``seekg'' member.
      // MergeBuffer are created for a ``ifstream''
      // but ``cin'' is also an ``istream''
      source->seekg(mergepointstart);
      source->read( charbuf+eun, toberead );
      // The following fails under MSDOS with text files.
      // We exclusively use binary files therefore
      assert( source->gcount() == toberead );
   }

   TRACE<<"Have read: "<<toberead<<'\n';
   mergepointstart += toberead;
   eun += toberead;

   allread = mergepointstart==mergepointend;
}


/*****************************************************************************/
/*                                                                           */
/*  Read the buffer full from ``mergepointstart'' in ``source'' and          */
/*  transform it into a sorted heap of SortString 's. Leftovers from the     */
/*  previous read are used up, before physically reading.                    */
/*                                                                           */
/*****************************************************************************/

void MergeBuffer::refill ( )
{
   TRACE<<"refilling from "<<mergepointstart<<"\n";
   read();
   sort();
   remainderstart = charbuf+bun;
   remainderlen   = eun-bun;
   TRACE<<"refilled up to "<<mergepointstart<<" eun ="<<eun<<"\n";

}
