/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : sbuffer.h
  DESCRIPTION : the smart input buffer of sorting program (definition)
  PROJECT : SSORT
  AUTHOR : A vd Horst
  LANGUAGE : Turbo C++ (3.0)
  CREATED : 920906 AH
  LAST CHANGE : 950220 AH enough_memory_for introduced
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950202 AH eof splitted into allread and eof
  LAST CHANGE : 950202 AH cleanup comment etc.
  LAST CHANGE : 921117 AH row destructor (delete [])
  LAST CHANGE : 921109 AH comment clean up
  LAST CHANGE : 921108 AH References to stream used
  LAST CHANGE : 921106 AH Eliminated estimated length parameter on creation
  LAST CHANGE : 921106 AH Introduced CHAR typedef
  LAST CHANGE : 920926 Merge/SortBuffer specialisations of Buffer
  LAST CHANGE : 920926 introduced inheritance from SortedHeap<SortString>
  LAST CHANGE :
                                                                   */
/*******************************************************************/

// requires #include <string.h>
// requires #include "loctypes.h"
// requires #include "sstring.h"

/*****************************************************************************/
/*                                                                           */
/* ``Buffer'' is a kind of smart buffer on the input stream ``source''.      */
/* Outside view: a bunch of records that can retrieved in sorted order,      */
/* replenished with ``refill''.                                              */
/* Inside :                                                                  */
/* It tries to read in as much as possible, triggered by ``refill'' (or      */
/* creation). It divides the input in records (class SortString) by          */
/* ``getsortstring'' ; ``sort'' puts them in a heap.                         */
/* These records are removed in sorted order from the heap by successive     */
/* calls of ``removeitem''                                                   */
/* The buffer cannot be released before all records are removed.             */
/*                                                                           */
/*****************************************************************************/

class Buffer : public SortedHeap< SortString >
{
public:
   istream *source;     // Input stream, already open
   CHAR *charbuf;       // Buffers input from ``source''
   int charbufsize;     // Usable length of ``charbuf''
   int  *lenbuf;        // Buffer for field lengths
   int  *plenbuf;       // First int in ``lenbuf'' not yet filled in
   Buffer(istream &is, int bl); // Create buffer on ``is'' with charbufsize=bl
   ~Buffer() {
      assert(nlines==0);
      delete [] lenbuf;
      delete [] charbuf;
      delete [] lines;
   };
   bool enough_memory(); // There is enough memory on the heap to
                         // accommodate another buffer of this length
   static bool Buffer::enough_memory_for( int len , int amount=1 );
                        // There is enough memory on the heap to accommodate 
                        // ``amount'' buffers of length ``len''
   void sort();         // Assign ``charbuf'' to strings in SortedHeap
   int bun;             // Start of characters read but unassigned
   int eun;             // One past characters read but unassigned

protected:
   bool allread;        // Everything from ``source'' readin,
                        // not necessarily transformed into SortString 's
private:
   bool nosep;                  // charbuf[bun..eun] doesnot contain a separator
   SortString getsortstring();  // Chips off one string from buf[bun..eun-1]
#ifdef DEBUG
   int  *lenbufend;     // To test against for overflow of ``lenbuf''
#endif
};


/*****************************************************************************/
/*                                                                           */
/*   Instantiation of ``Buffer'' with following narrowing down:              */
/*   -the last character to be read is the last character of the file        */
/*   -all calls of SortBuffer apply to the same istream and the stream must  */
/*    remain open. It is not possible to restart a new set of buffers with   */
/*    the same, or some other, istream.                                      */
/*   -``remainder'' is associated with ``in'', hence fixed.                  */
/*                                                                           */
/*****************************************************************************/

class SortBuffer : public Buffer
{
public:
   SortBuffer(istream &is, int buflen=32000 );
   SortString getsortstring();
   static CHAR *remainderstart; // Holds characters left over from
   static int remainderlen;     //   previous SortBuffer creation
   void read();
   void refill();
   bool eof;            // ``source'' is totally consumed
};


/*****************************************************************************/
/*                                                                           */
/*   The following class is necessary, just to be able to redefine           */
/*   '<' for pointers ``to SortBuffer'' (because '<' has an intrinsic        */
/*   meaning for pointers). We need '<' to put those items in a              */
/*   SortedHeap themselves.                                                  */
/*/                                                                          */
/*                                                                           */
/*****************************************************************************/

class pSortBuffer
{
public:
   SortBuffer *sb;

   bool operator< ( pSortBuffer psb2 )
   {
      return sb->lines[0] < psb2.sb->lines[0];
   }
};


/*****************************************************************************/
/*                                                                           */
/*   ``MergeBuffer'' is a kind of smart buffer on the input.                 */
/*   It is an instantiation of ``Buffer'' with following narrowing down:     */
/*   -the characters to be read are from the file ``ifstream'' from the      */
/*   initial position of ``mergepointstart'' up to ``mergepointend''.        */
/*   -``remainder'' is associated with the above part of ``istream''         */
/*   -all calls of ``refill'' apply to that part.                            */
/*   Each call of ``refill'' boosts ``mergepointstart''                      */
/*                                                                           */
/*****************************************************************************/

class MergeBuffer : public Buffer
{
private:
   long mergepointstart;    // Between these two points is the sorted
   long mergepointend;      //  part of the file, that has not been merged yet.
   void read();
   CHAR *remainderstart; // Holds characters left over from
   int remainderlen;     //   previous SortBuffer creation
public:
   MergeBuffer(                 // Create a merge buffer on ``in'' for the
      ifstream &is,             // part of the file  mergepoint[0]
      long *mergepoint,         // mergepoint[1], with a buffer of
      int len=32000             // lenght ``len''
   );
   void refill();               // replenish from the assigned file part
};


/*****************************************************************************/
/*                                                                           */
/*   The following class is necessary, just to be able to redefine           */
/*   '<' for pointers ``to MergeBuffer''. See als pSortBuffer.               */
/*                                                                           */
/*****************************************************************************/

class pMergeBuffer
{
public:
   MergeBuffer *sb;

   bool operator< ( pMergeBuffer psb2 )
   {
      return sb->lines[0] < psb2.sb->lines[0];
   }
};
