/*******************************************************************/
/*
  COPYRIGHT: (c) 1992@2000 by SP&AR&C HOLLAND by GNU Public Licence 
  FILE : sheap.h
  DESCRIPTION : heapsort in behalf of sorting program (template)
  PROJECT : SSORT
  AUTOR : A vd Horst
  LANGUAGE : Turbo C++ (2.0)
  CREATED : 920905 AH
  LAST CHANGE : 950206 AH total clean up, configuration
  LAST CHANGE : 950203 heapsize made private
  LAST CHANGE : 950129 Full introduced
  LAST CHANGE : 921214 tracing requires iostream.h
  LAST CHANGE : 921213 error message changed into assert
  LAST CHANGE : 921106 heap generated with length 0 has no storage
  LAST CHANGE : 920908 pointer type for Item no longer obligatory
*/

// requires <iostream.h>
// requires <assert.h>

/*****************************************************************************/
/*                                                                           */
/*   A ``SortedHeap'' is used to build a sorted heap by successively adding  */
/*   Items. Removing takes place in sorted order. It is required that        */
/*   the operator '<' works for the class ``Item''.                          */
/*   All public methods have the pre and postcondition that it is a proper   */
/*   sorted heap.                                                            */
/*   Definition of a sorted heap :                                           */
/*    lines[] is sorted such that for all i < nlines:                        */
/*       lines[i] < lines[2*i+1] && lines[i] < lines[2*i+2]                  */
/*   See also Knuth Art of programming: Sorting and Searching.               */
/*   Note the difference, we are counting from 0.                            */
/*   Note for the not mathematically aware:                                  */
/*   the operator> is extended to work between Items and sets of             */
/*   Items, by the definition that a set is greater iff all of its elements  */
/*   are. For heaps the operator> is trivial to implement, the operator< is  */
/*   not(!). It is also not true that either i<=s or i>=s for i an Item      */
/*   and s a set.                                                            */
/*                                                                           */
/*****************************************************************************/

template <class Item>
class SortedHeap
{
public:

   // Generate an empty heap with a maximum of len entry's
   SortedHeap(int len) :
      lines( new Item[heapsize=min(len,0xff00U/sizeof(Item))] )
   {
      TRACE<<"heapsize "<<heapsize<<'\n';
      assert(lines);
      nlines = 0;
   };

   // Put the item in the heap
   void additem( Item );

   // Remove lowest item from heap
   Item removeitem( );

   // Indicate no more item's present
   bool empty() { return nlines == 0; };

   // Indicate "no more item's could be added"
   bool full() { return nlines == heapsize; };

   // An item is smaller than a heap, iff it is smaller than all elements
   // This defines "an heap is greater than an item", because of syntactic
   // problems.
   bool operator> ( Item item )
   {
      assert( !empty() );
      return item<lines[0] ;   // Still require only '<' between Item's
   };

   // Similar to operator>
   bool operator>= ( Item item )
   {
      assert( !empty() );
      return  ! (lines[0]<item) ;
   };

   // SortedHeaps compare as do their lowest elements
   bool operator< ( SortedHeap<Item> sh2 )
   {
      assert( !empty() ); assert( !sh2.empty() );
      return lines[0] < sh2.lines[0];
   };

   int nlines;    // Valid # lines in ``lines''
   Item *lines;   // lines[0..nlines-1] form a heap with respect to '<'

protected:
   int heapsize;                // Maximum number of lines present

private:
   void adjust( int nline );
   void fill_hole( int nline );
};

/*******************************************************************/
/*
  DESCRIPTION : heapsort in behalf of sorting program (implementation)
  CREATED : 920905 AH
  LAST CHANGE :
*/
/*******************************************************************/

// REMARK : The implementation cannot be put in a spearate module.
//          TURBO C++ simply would not generate actual code

/*****************************************************************************/
/*                                                                           */
/*   ``adjust'' makes sure that a newly added line with index ``newline''    */
/*   is adjusted to a place that the array ``lines'' is again a heap.        */
/*                                                                           */
/*****************************************************************************/

template <class Item>
void SortedHeap<Item>::adjust( int newline )
{
   Item swap;

   while ( newline != 0 )
   {
      int boss = (newline+1)/2 - 1 ;

      // If this "boss" is not a boss : swap
      if ( lines[newline]<lines[boss] )
      {
         swap = lines[boss];
         lines[boss] = lines[newline];
         lines[newline] = swap;
      }
      newline = boss;  // In fact recursive : adjust(boss)
   }
}

/*****************************************************************************/
/*                                                                           */
/*   ``fillhole'' makes sure that a hole with index ``newline'' is filled    */
/*   such that the array ``lines'' is again a heap.                          */
/*   Before : heap is valid except w.r.t. entry ``hole''                     */
/*   After : heap is valid with entry ``hole'' less                          */
/*                                                                           */
/*****************************************************************************/

template <class Item>
void SortedHeap<Item>::fill_hole( int hole )
{
   int isubo1;       // Subordinate number 1

   while ( (isubo1 = 2*hole+1) < nlines )     // Is there a subordinate?
   {
      if ( isubo1 == nlines-1 )
      {
         // One subordinate, nothing to choose
         lines[hole] = lines[isubo1];
         hole = isubo1;         // fill the new hole
      }
      else     // Two subordinates, choose lowest
      {
         int lowest =
           ( lines[isubo1]<lines[isubo1+1] ? isubo1 : isubo1+1 );
         lines[hole] = lines[lowest];
         hole = lowest;         // fill the new hole
      }
   }

   lines[hole] = lines[--nlines];// Plug hole with last line
   adjust(hole);     // which can destroy the heap
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

template <class Item>
void SortedHeap<Item>::additem(Item s)
{
   assert( nlines < heapsize );
   lines[nlines++] = s;
   adjust( nlines-1 );
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

template <class Item>
Item SortedHeap<Item>::removeitem()
{
   Item lowest;
   if ( nlines == 0 )
   {
      cerr<<"nlines "<< nlines<<"\n";
      assert( nlines!=0 );
   }

   lowest = lines[0];        // Property of heap
   fill_hole( 0 );
   return lowest;
}
